# Brand new vault
from UtilsRPE.Functions import load_vault
from collector import SSH
import re
import time
import sys

# Signal test
import signal


def hash_from_list(_list):
    _list = list(sorted(_list))
    val = (sys.maxsize + 1) * 2
    return hash(tuple(_list)) + val


class SimulationError(Exception):
    pass


def generate_jobscript(template, parameters):
    with open(template) as f:
        job = f.read()
        for parameter, value in parameters.items():
            job = job.replace("%%%s%%" % parameter, value)
    return job


def submit_job(job, remote_jobscript, communicator):
        with communicator.sftp.file(remote_jobscript, "w") as jobscript:
            jobscript.write(job)

        command = "sbatch %s" % remote_jobscript
        stdin, stdout, stderr = communicator.execute(command)
        output = str(stdout.read())

        jobid_pattern = "Submitted batch job +([0-9]+)"
        jobid = re.search(jobid_pattern, output).group(1)
        return jobid


def job_status(job_id, communicator):
    command = 'sacct --job %s --format="State,Elapsed"' % job_id
    stdin, stdout, stderr = communicator.execute(command)
    output = stdout.read().decode()
    try:
        status = output.split("\n")[2].split()[0]
        time = output.split("\n")[2].split()[1]
    except IndexError:
        status = "PENDING"
        time = "00:00"
    return status, time


def obtain_result(remote_rundir, communicator):
    with communicator.sftp.open("%s/accuracy.txt" % remote_rundir) as f:
        error = float(f.read())
        return error


def generate_namelist(variable_set, filename, _vault=None):
    if _vault:
        vault = _vault
        variables = list(sorted([v for v in vault.variables if v.id is not None], key=lambda x: -x.error))
    for index, v in enumerate(variables):
        if v.id in variable_set:
            v.precision = 10
        else:
            v.precision = 52
    vault.generate_namelist_precisions(filename=filename, exclude_double_precision=True)


def signal_handler(sig, frame):
    jobs_to_cancel = " ".join(job_ids)
    print('\n\nCanceling simulation/s %s !!' % jobs_to_cancel)
    command = "scancel %s" % jobs_to_cancel
    stdin, stdout, stderr = remote.execute(command)
    print(stdout.read().decode())
    exit(1)


class Job:
    def __init__(self, variable_set, remote_communicator, vault=None):
        # Init variables
        self.vault = vault
        self.variable_set = variable_set
        self.communicator = remote_communicator
        self.remote_scratch = self.communicator.remote_scratch
        self.remote_logdir = "%s/LOGS" % self.remote_scratch
        self.remote_namelist_dir = "%s/Namelists" % self.remote_logdir

        self.remote_rundir = "%s/RefinementTest%i" % (self.remote_scratch, self.hash)

        self.remote_jobscript_path = "%s/refinement_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/refinement_namelist_%i" % (self.remote_namelist_dir, self.hash)
        self.local_namelist = "NewWorlds/namelist_precisions"

        self.template = "Templates/singleJobscript.sh"
        # parameters to replace:
        self.parameters = {"JOBNAME": "refinement_test",
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }

        self.jobscript = generate_jobscript(self.template, self.parameters)

        self.is_completed = False
        self.is_successful = None
        self.elapsed_time = None
        self.job_id = None
        self._result = None
        self.submitted = False

    @property
    def status(self):
        if self.submitted is False:
            return "NOT SUBMITTED", "00:00:00"
        elif self.is_completed is False:
            _status, _time = job_status(self.job_id, self.communicator)
            _status = _status.strip()
            if _status == "COMPLETED":
                # self.is_completed = True
                self.elapsed_time = _time
            return _status, _time
        else:
            return "COMPLETED", self.elapsed_time

    # @property
    def result(self, level=0):
        if self._result is None:
            if level > 10:
                return None
            try:
                self._result = obtain_result(self.remote_rundir, self.communicator)
                return self._result
            except FileNotFoundError:
                # time.sleep(1)
                # return self.result(level=level+1)
                return None
        else:
            return self._result

    @property
    def hash(self):
        return hash_from_list(self.variable_set)

    def check_if_result_already_exists(self):
        self.update_parameters()
        accuracy_filename = "%s/accuracy.txt" % self.remote_rundir
        try:
            self.communicator.sftp.open(accuracy_filename)
            # self.is_completed = True
            self.submitted = True
            # Recovering job_id
            job_id_filename = "%s/job_id.txt" % self.remote_rundir
            self.job_id = self.communicator.sftp.open(job_id_filename).read().decode().strip()
            print("Job %s already finished! Error: %.2e" % (self.job_id, self.result()))
            return True
        except IOError:
            return False

    def submit(self):
        self.submitted = True
        # Check if result already exists
        completed = self.check_if_result_already_exists()
        if not completed:
            self.update_parameters()
            # Generate namelist and put it in the remote system
            generate_namelist(self.variable_set, self.local_namelist, self.vault)
            self.communicator.put(self.local_namelist, self.remote_namelist)
            # Submit job and obtain job id
            self.job_id = submit_job(self.jobscript, self.remote_jobscript_path, self.communicator)
            print("Job %i successfully submitted to the remote machine with job_id %s" % (self.hash, self.job_id))
        else:
            # self.is_completed = True
            pass

    def cancel_remote(self):
        if self.job_id is not None:
            command = "scancel -v %s" % self.job_id
            print("Canceling job %s" % self.job_id)
            stdin, stdout, stderr = self.communicator.execute(command)
            print(stdout.read().decode())
            print(stderr.read().decode())
            self.job_id = None

    def update_parameters(self):
        self.remote_rundir = "%s/RefinementTest%i" % (self.remote_scratch, self.hash)

        self.remote_jobscript_path = "%s/refinement_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/refinement_namelist_%i" % (self.remote_namelist_dir, self.hash)
        self.local_namelist = "NewWorlds/namelist_precisions_%i" % self.hash

        # parameters to replace:
        self.parameters = {"JOBNAME": "refinement_test",
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }
        self.jobscript = generate_jobscript(self.template, self.parameters)


def create_case(variable_set):

    # Remote parameters
    remote_scratch = "/gpfs/scratch/pr1egi00/pr1egi23/ROMS-runs/"

    return Job(variable_set, remote_scratch)


def equal_0(index, var):
    if var.error == 0:
        return True
    else:
        return False


def less_than_threshold(index, var, _threshold):
    if var.error < _threshold:
        return True
    else:
        return False


def less_than_number(index, var, _number):
    if index > _number:
        return True
    else:
        return False


def generate_set(limit):
    return [v.id for i, v in enumerate(variables) if less_than_number(i, v, limit) and less_than_threshold(i, v, 10 ** -2)]


def set_difference(job1, job2):
    diff1, diff2 = set_differences(job1, job2)
    return len(diff1+diff2)


def set_differences(job1, job2):
    # assert isinstance(job1, Job) and isinstance(job2, Job)
    set1 = job1.variable_set
    set2 = job2.variable_set
    diff1 = [v for v in set1 if v not in set2]
    diff2 = [v for v in set2 if v not in set1]
    return diff1, diff2


def set_1_contained_in_2(job1, job2):
    set1 = job1.variable_set
    set2 = job2.variable_set
    diff1 = [v for v in set1 if v not in set2]
    if len(diff1) == 0:
        return True
    else:
        return False


# Main PROGRAM
if __name__ == "__main__":
    # Establish SSH connection to remote machine
    hostname = "mn1.bsc.es"
    user = "pr1egi23"
    remote = SSH(user=user, host=hostname)
    max_running_jobs = 10

    # Load variable database
    vault_path = "vault_with_processed_a1bo.pkl"
    vault = load_vault(vault_path)

    variables = sorted([v for v in vault.variables if v.id is not None], key=lambda x: -x.error)

    # Initialize two empty lists for running and completed experiments
    experiments = []
    running_jobs = []

    case_type = "regroup"
    if case_type == "incremental":
        # Generate different cases (Incremental Approach)
        group_size = 10
        # Finding number of variables to be included
        top = len(variables)
        for i, v in enumerate(variables):
            if v.error == 0:
                top = i
                break
        limits = list(reversed(range(0, top, group_size)))
        all_sets = [generate_set(l) for l in limits]
    elif case_type == "regroup":
        # Generate different cases (Re-grouping approach)
        group_size = 10
        # Finding number of variables to be included
        threshold = 10 ** -2
        groupable_variables = [v.id for v in variables if v.error < threshold]

        all_sets = []
        while len(groupable_variables):
            current_group = []
            for i in range(group_size):
                current_group.append(groupable_variables.pop(0))
                if len(groupable_variables) == 0:
                    break
            all_sets.append(current_group)
    else:
        raise AssertionError("Unknown case_type")

    sets = all_sets

    # Setting a signal trap to cancel running experiments if the program is stopped
    signal.signal(signal.SIGINT, ºsignal_handler)

    all_jobs = [create_case(set_t) for set_t in sets]

    # Check if jobs have been already ran
    for job in all_jobs:
        job.check_if_result_already_exists()

    pending_jobs = [job for job in all_jobs if job.is_completed is False]
    completed_jobs = [job for job in all_jobs if job.is_completed is True]
    suspended_jobs = []

    # for job in completed_jobs:
    #     if job.result > threshold:
    #         experiments_above_threshold.append(job)


    def find_frontier(experiments_above_threshold):
        # Look for the failed experiment most similar to a successful one
        min_diff = 1000
        first_above = None
        last_below = None
        for job in experiments_above_threshold:
            difference = 1000
            result = None
            for job2 in completed_jobs:
                if job2 is not job:
                    if job2.result() < threshold:
                        if set_difference(job, job2) < difference:
                            result = job2
                            difference = set_difference(job, job2)

            if difference < min_diff:
                min_diff = difference
                first_above = job
                last_below = result
        return first_above, last_below


    for job in completed_jobs:
            cancel_jobs = False
            find_most_similar = False

            if cancel_jobs:
                for job2 in pending_jobs:
                    if set_1_contained_in_2(job, job2):
                        print("\t%i" % job2.hash)
                        suspended_jobs.append(job2)
                        pending_jobs = [j for j in pending_jobs if j != job2]
            if find_most_similar:
                difference = 1000
                result = None
                for job2 in completed_jobs:
                    if job2 is not job:
                        if job2.result() < threshold:
                            if set_difference(job, job2) < difference:
                                result = job2
                print(set_difference(job, result))

                sensitive_variables = set_differences(job, result)[0]
                for variable in sensitive_variables:
                    new_set = result.variable_set + [variable]
                print(sensitive_variables)

    Run = False
    if Run:

        # While there are still experiments to run
        while len(pending_jobs) > 0:

            # Submission part
            # Submit pending experiments until reaching maxim number of running experiments
            while len(running_jobs) < max_running_jobs:
                job = pending_jobs.pop(0)
                # Submit job
                job.submit()
                # Append job to running experiments list
                running_jobs.append(job)
                experiments.append(job)

            # Obtain list of ids corresponding to running jobs
            # (used to suspend running jobs if the program is canceled).
            job_ids = [j.job_id for j in running_jobs if j.job_id is not None]

            # Check running experiment status
            for index, job in enumerate(running_jobs):
                status, runtime = job.status
                print("ID: %s Status: %s ElapsedTime: %s" % (job.job_id, status, runtime))
                if job.is_completed:
                    completed_jobs.append(running_jobs.pop(index))

            sleep_time = 10
            time.sleep(sleep_time)

        while len(running_jobs) > 0:
            for index, job in enumerate(running_jobs):
                status, runtime = job.status
                print("ID: %s Status: %s ElapsedTime: %s" % (job.job_id, status, runtime))
                if job.is_completed:
                    completed_jobs.append(running_jobs.pop(index))

        print(len(completed_jobs))
        for job in completed_jobs:
            print("Job with id %s finished with error %.2e" % (job.job_id, job.result()))

